# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.log
:synopsis: Logging facilities for this project.
:author: Amr Ali <amr@databracket.com>

.. autoclass:: Logger
    :members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import logging
import logging.handlers

from bastio.mixin import public, UniqueSingletonMeta

def _getSyslogAddress():
    import sys

    if sys.platform.startswith('linux'):
        return '/dev/log'
    elif sys.platform.startswith('darwin'):
        return '/var/run/syslog'

@public
class Logger(object):
    """A class that provides logging facilities exposed by the underlying
    Python's logging module.
    """
    __metaclass__ = UniqueSingletonMeta

    def __init__(self, prog = 'bastio-agent'):
        self._logger = logging.getLogger(prog)
        self._formatter = logging.Formatter(prog + '[%(process)d]: %(message)s')
        self._logger.addHandler(logging.NullHandler())
        self._syslog_enabled = False
        self._stream_enabled = False

    def __getattr__(self, name):
        return getattr(self._logger, name)

    def enable_syslog(self):
        """Enable Syslog logging. This method is idempotent."""
        if self._syslog_enabled:
            return
        _handler = logging.handlers.SysLogHandler(address=_getSyslogAddress(),
                facility=logging.handlers.SysLogHandler.LOG_SYSLOG)
        _handler.setFormatter(self._formatter)
        self._logger.addHandler(_handler)
        self._syslog_enabled = True

    def enable_stream(self):
        """Enable logging to STDOUT. This method is idempotent."""
        if self._stream_enabled:
            return
        _handler = logging.StreamHandler()
        _handler.setFormatter(self._formatter)
        self._logger.addHandler(_handler)
        self._stream_enabled = True




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.configs
:synopsis: Configurations memory and file system store.
:author: Amr Ali <amr@databracket.com>

.. autoclass:: GlobalConfigStore
    :members:
    :special-members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import os
import ConfigParser

from collections import defaultdict

from bastio.mixin import public, UniqueSingletonMeta
from bastio.excepts import BastioConfigError, reraise

@public
class GlobalConfigStore(defaultdict):
    """An in-memory configuration store that inherits :class:`defaultdict`
    behavior. This store is a singleton so only one object is created.
    Please see :class:`bastio.mixin.UniqueSingletonMeta`.

    The in-memory store takes precedence over the configuration file. If key
    was not found in the memory store we'll try to get it from the configuration
    file, if all fails we will simply return the default value of the
    ``default_factory`` assigned in the constructor.
    """
    __metaclass__ = UniqueSingletonMeta

    def __init__(self, section='agent', default_factory=str):
        """
        :param section:
            The section of the configuration file to use by default.
        :type section:
            str
        :param default_factory:
            See :class:`collections.defaultdict`
        :type default_factory:
            object
        """
        super(GlobalConfigStore, self).__init__(default_factory)
        self.__dict__['_section'] = section
        self.__dict__['_config'] = None

    def __getattr__(self, attr):
        """
        There's only one case where precedence is reversed in favor of the
        configuration file; when one of the [method]s below are present right
        before the <option name>. This and once the value is retrieved successfully
        it will be set in the memory store so that you can access it later without
        having to consult the configuration file again.

        :param attr:
            The syntax is [method]_[section]_<option name> where [method] and
            [section] are optional, and <option name> is required. Note that
            if you wish not to supply [method] you need to remove [section] and
            ``_`` as well. Also if [section] was not supplied the default section
            name that was supplied to the constructor will be used.

            [method] could be one of a few possibilities. Either ``get``,
            ``getint``, ``getfloat``, or ``getboolean`` that correspond to the
            methods available through the :class:`ConfigParser` interface.

            e.g., ``get_alpha_name`` will try to get the option ``name`` from
            the configuration file, similarly in the case of ``get_name`` we will
            try to get the option's value from the configuration file but from
            under the section that was supplied to the constructor.
        :type attr:
            str
        :returns:
            The option's value or ``default_factory`` if no value was set.
        :raises:
            :class:`bastio.excepts.BastioConfigError`
        """
        if attr.startswith('get'):
            tmp = attr.split('_')
            method = tmp[0]

            # Check if config parser is loaded and ready
            if not self._config:
                raise BastioConfigError('no configuration file was loaded')

            # Check if the method requested is supported
            if not hasattr(self._config, method):
                raise BastioConfigError('method `{}` is not supported'.format(
                    method))

            # Check if <option name> was supplied
            if len(tmp) < 2:
                raise BastioConfigError(
                        'must supply an option name after the method `{}`'.format(
                            method))

            # Syntax is now assumed to be [method]_<option name>
            if len(tmp) == 2:
                method = tmp[0]
                option = '_'.join(tmp[1:])
                try:
                    self[option] = getattr(self._config, method)(self._section,
                            option)
                except ConfigParser.NoOptionError:
                    return self[option]
                except ConfigParser.Error:
                    reraise(BastioConfigError)
                return self[option]

            # Handle the case where [section] is also provided
            if len(tmp) > 2:
                method = tmp[0]
                section = tmp[1]
                option = '_'.join(tmp[2:]) # Treat the rest as the option name
                try:
                    self[option] = getattr(self._config, method)(section, option)
                except ConfigParser.NoOptionError:
                    return self[option]
                except ConfigParser.Error:
                    reraise(BastioConfigError)
                return self[option]
        else:
            if self._config:
                # Check if it exists in the memory store and if not try the
                # configuration file
                if attr in self:
                    return self[attr]

                # Now let's try the configuration file
                try:
                    self[attr] = self._config.get(self._section, attr)
                except ConfigParser.NoOptionError:
                    # Not found here either, return value of ``default_factory``
                    return self[attr]
                except ConfigParser.Error:
                    reraise(BastioConfigError)
            return self[attr]

    def __setattr__(self, attr, value):
        """Set a value to the in-memory store.

        :param attr:
            The key in the key-value in-memory store.
        :type attr:
            str
        :param value:
            The value in the key-value in-memory store.
        :type value:
            object
        """
        self[attr] = value

    def load(self, filename):
        """Load a configuration file using :class:`ConfigParser` into memory.

        :param filename:
            The absolute path of the configuration file.
        :type filename:
            str
        :raises:
            :class:`bastio.excepts.BastioConfigError`
        """
        if not os.path.exists(filename):
            raise BastioConfigError(
                    'configuration file `{}` does not exist'.format(filename))
        self.__dict__['_config'] = ConfigParser.SafeConfigParser()
        ret = self._config.read(filename)
        if not ret:
            raise BastioConfigError(
                    'could not load configuration file: `{}`'.format(filename))




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.test.test_mixin
:synopsis: Unit tests for the mixin module.
:author: Amr Ali <amr@databracket.com>
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import unittest
from bastio.mixin import KindSingletonMeta, UniqueSingletonMeta, Json

class KindTmpClass(object):
    __metaclass__ = KindSingletonMeta

    def __init__(self, name):
        self.name = name

class UniqueTmpClass(object):
    __metaclass__ = UniqueSingletonMeta

    def __init__(self, name):
        self.name = name

class TestSingleton(unittest.TestCase):
    def test_kind_singleton(self):
        test1 = KindTmpClass("test1")
        test2 = KindTmpClass("test2")
        self.assertEqual(test1.name, "test1")
        self.assertEqual(test2.name, test1.name)
        self.assertEqual(test1, test2)

    def test_unique_singleton(self):
        test1 = UniqueTmpClass("test1")
        test2 = UniqueTmpClass("test2")
        self.assertEqual(test1.name, "test1")
        self.assertEqual(test2.name, "test2")
        self.assertNotEqual(test2.name, test1.name)
        self.assertNotEqual(test1, test2)

class TestJson(unittest.TestCase):
    def test_json_mixin(self):
        obj = Json()
        obj.field_str = 'test'
        obj.field_int = 434
        obj.field_float = 4.34
        obj.field_sub = Json()
        obj.field_sub.field_str = 'sub_test'
        obj.field_sub.field_int = 123
        obj.field_sub.field_float = 1.23
        obj_json = Json().from_json(obj.to_json())
        self.assertEqual(obj_json.to_json(), obj.to_json())
        self.assertIn('field_str', obj_json)
        self.assertIsInstance(obj_json.field_str, basestring)
        self.assertIsInstance(obj_json.field_int, int)
        self.assertIsInstance(obj_json.field_float, float)
        self.assertIsInstance(obj_json.field_sub, Json)

        obj_json.field_fail = [Json()]
        with self.assertRaises(TypeError):
            obj_json.to_json()

tests = [
        TestSingleton,
        TestJson,
        ]




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.ssh
:synopsis: A Package for SSH management and communication.
:author: Amr Ali <amr@databracket.com>

.. rst-class:: html-toggle

Backend Connection Management
-----------------------------
.. automodule:: bastio.ssh.client

.. rst-class:: html-toggle

Action Processor
----------------
.. automodule:: bastio.ssh.api

.. rst-class:: html-toggle

Protocol Parsing And Logic Routing
----------------------------------
.. automodule:: bastio.ssh.protocol

.. rst-class:: html-toggle

Cryptographic Utilities
-----------------------
.. automodule:: bastio.ssh.crypto
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.excepts
:synopsis: Exception classes used by this project.
:author: Amr Ali <amr@databracket.com>

.. autofunction:: reraise

.. autoexception:: BastioException

.. autoexception:: BastioConfigError

.. autoexception:: BastioUnimplementedError

.. autoexception:: BastioTaskError

.. autoexception:: BastioCryptoError

.. autoexception:: BastioNetstringError

.. autoexception:: BastioEOFError

.. autoexception:: BastioMessageError

.. autoexception:: BastioBackendError

.. autoexception:: BastioAccountError
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import sys

from bastio.mixin import public

@public
def reraise(ex, message=None):
    """Reraise the exception last happened with the original traceback.

    :param ex:
        The exception to be raised instead of the original one.
    :type ex:
        A subclass of Exception.
    :param message:
        An optional message to replace the old exception's message.
    :type message:
        str
    :raises:
        ``ex``
    """
    (_, msg, tb) = sys.exc_info()
    msg = message if message else msg
    raise ex, msg, tb

@public
class BastioException(Exception):
    """A Bastio general exception"""
    pass

@public
class BastioConfigError(BastioException):
    """A configuration operation error"""
    pass

@public
class BastioUnimplementedError(BastioException):
    """An unimplemented operation error"""
    pass

@public
class BastioTaskError(BastioException):
    """A task operation error"""
    pass

@public
class BastioCryptoError(BastioException):
    """A cryptographic operation error"""
    pass

@public
class BastioNetstringError(BastioException):
    """A Netstring parsing or composition error"""
    pass

@public
class BastioEOFError(BastioException):
    """EOF error"""
    pass

@public
class BastioMessageError(BastioException):
    """Message parser operation error"""
    pass

@public
class BastioBackendError(BastioException):
    """A backend client operation error"""
    pass

@public
class BastioAccountError(BastioException):
    """An account operation error"""
    pass




# Copyright 2013 Databracket LLC.
# See LICENSE file for details.

"""
:module: bastio.version
:synopsis: A module that contain version details for the package.
:author: Amr Ali <amr@databracket.com>

Update version information here only. There's no need to update any other
file in the package.

Release of a hotfix/bugfix:
- Increment version's ``BUILD`` field.

Release of an update that modifies the API but maintains
backward compatibility:
- Increment version's ``MINOR`` field and reset the ``BUILD`` field.

Release of an update that that modifies the API and breaks
backward compatibility:
- Increment version's ``MAJOR`` field and reset both the ``MINOR`` and ``BUILD`` fields.
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

VERSION_MAJOR = 0
VERSION_MINOR = 1
VERSION_BUILD = 1

# NOTE: DO NOT MODIFY ANYTHING BELOW THIS LINE UNLESS YOU DO KNOW WHAT YOU ARE DOING.

VERSION_INFO = (VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD)
VERSION_STRING = '%d.%d.%d' % VERSION_INFO

__version__ = VERSION_STRING




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.ssh.crypto
:synopsis: A module for miscellaneous cryptographic facilities.
:author: Amr Ali <amr@databracket.com>

.. autoclass:: RSAKey
    :members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import base64
import StringIO
import paramiko

from bastio.mixin import public
from bastio.excepts import BastioCryptoError, reraise

@public
class RSAKey(paramiko.RSAKey):
    """A class to add a few helper functions to :class:`paramiko.RSAKey`."""

    @classmethod
    def generate(cls, size):
        """A wrapper around paramiko's RSAKey generation interface to add this
        class's methods to it.

        :param size:
            The size of the RSA key pair in bits you wish to generate.
        :type size:
            int
        :returns:
            :class:`RSAKey`
        """
        try:
            klass = paramiko.RSAKey.generate(size)
            klass.__class__ = cls
            return klass
        except Exception:
            reraise(BastioCryptoError)

    @classmethod
    def from_public_key(cls, public_key):
        """Load an instance of this class with ``public_key``.

        :param public_key:
            An OpenSSH formatted public key.
        :type public_key:
            str
        :returns:
            :class:`RSAKey`
        """
        if not cls.validate_public_key(public_key):
            raise BastioCryptoError("invalid public key")
        vals = public_key.split(' ')
        return cls(data=base64.decodestring(vals[1]))

    @classmethod
    def validate_public_key(cls, data):
        """Validate public key data.

        :param data:
            An OpenSSH formatted public key.
            [ssh-][ALGO][ ][BASE64][ ][COMMENT]
        :type data:
            str
        :returns:
            Whether the key data is valid.
        """
        try:
            if not data.startswith('ssh-'):
                return False
            vals = data.split(' ')
            if len(vals) < 2: # There must be at least 'ssh-algo' and 'base64'
                return False
            cls(data=base64.decodestring(vals[1]))
            return True
        except Exception:
            return False

    @classmethod
    def validate_private_key(cls, data):
        """Validate private key data.

        :param data:
            A PEM formatted private key content.
        :type data:
            str
        :returns:
            Whether the key data is valid.
        """
        try:
            buf = StringIO.StringIO(data)
            cls.from_private_key(buf)
            return True
        except Exception:
            return False

    @classmethod
    def validate_private_key_file(cls, filename):
        """Validate private key file.

        :param filename:
            A filename that contains PEM formatted private key data.
        :type filename:
            str
        :returns:
            Whether the key file is valid.
        """
        try:
            cls.from_private_key_file(filename)
            return True
        except Exception:
            return False

    def get_private_key(self):
        """Return private key data in PEM format."""
        buf = StringIO.StringIO()
        self.write_private_key(buf)
        buf.seek(0)
        return buf.read()

    def get_public_key(self):
        """Return the public key in OpenSSH authorized_keys line format."""
        return self.get_name() + ' ' + self.get_base64()




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.test.test_ssh_api
:synopsis: Unit tests for the ssh.api module.
:author: Amr Ali <amr@databracket.com>
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import os
import unittest

from bastio.ssh.api import Processor
from bastio.ssh.crypto import RSAKey
from bastio.ssh.protocol import (FeedbackMessage, AddUserMessage,
        RemoveUserMessage, UpdateUserMessage, AddKeyMessage, RemoveKeyMessage)
from bastio.concurrency import GlobalThreadPool

@unittest.skipIf(os.getuid() != 0, "this test case requires root access")
class TestProcessor(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._public_key = RSAKey.generate(1024).get_public_key()
        cls._proc = Processor()

    @classmethod
    def tearDownClass(cls):
        cls._proc.stop()

    def setUp(self):
        self._add_user(FeedbackMessage.SUCCESS, sudo=False)

    def tearDown(self):
        self._remove_user(FeedbackMessage.SUCCESS)

    def test_add_remove_user(self):
        self._add_user(FeedbackMessage.INFO, sudo=False)
        self._remove_user(FeedbackMessage.SUCCESS)
        self._remove_user(FeedbackMessage.INFO)
        self._add_user(FeedbackMessage.SUCCESS, sudo=False)

    def test_add_remove_key(self):
        self._remove_user(FeedbackMessage.SUCCESS)
        self._add_key(FeedbackMessage.ERROR)
        self._add_user(FeedbackMessage.SUCCESS, sudo=False)
        self._add_key(FeedbackMessage.SUCCESS)
        self._add_key(FeedbackMessage.INFO)
        self._remove_key(FeedbackMessage.SUCCESS)
        self._remove_key(FeedbackMessage.INFO)

    def test_update_user(self):
        self._remove_user(FeedbackMessage.SUCCESS)
        self._update_user(FeedbackMessage.ERROR, sudo=False)
        self._add_user(FeedbackMessage.SUCCESS, sudo=True)
        self._update_user(FeedbackMessage.SUCCESS, sudo=True)
        self._update_user(FeedbackMessage.SUCCESS, sudo=False)
        self._update_user(FeedbackMessage.ERROR, sudo=False)

    def _add_user(self, expect_status, **kwargs):
        msg = AddUserMessage(username="test_user", **kwargs)
        self._proc_message(msg, expect_status)

    def _remove_user(self, expect_status, **kwargs):
        msg = RemoveUserMessage(username="test_user", **kwargs)
        self._proc_message(msg, expect_status)

    def _update_user(self, expect_status, **kwargs):
        msg = UpdateUserMessage(username="test_user", **kwargs)
        self._proc_message(msg, expect_status)

    def _add_key(self, expect_status, **kwargs):
        msg = AddKeyMessage(username="test_user", public_key=self._public_key,
                **kwargs)
        self._proc_message(msg, expect_status)

    def _remove_key(self, expect_status, **kwargs):
        msg = RemoveKeyMessage(username="test_user", public_key=self._public_key,
                **kwargs)
        self._proc_message(msg, expect_status)

    def _proc_message(self, msg, expect_status):
        fb = self._proc.process(msg)
        self._assert_feedback(fb, expect_status)

    def _assert_feedback(self, fb, expect_status):
        self.assertIsInstance(fb, FeedbackMessage)
        msg = "{0} != {1}: {2}".format(fb.status, expect_status, fb.feedback)
        self.assertEqual(fb.status, expect_status, msg=msg)

tests = [
        TestProcessor,
        ]




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.ssh.protocol
:synopsis: A module containing protocol messages.
:author: Amr Ali <amr@databracket.com>

.. rst-class:: html-toggle

Low Level Protocols
-------------------
.. autoclass:: Netstring
    :members:

.. rst-class:: html-toggle

Message Bases
-------------
.. autoclass:: ProtocolMessage
    :members:

.. autoclass:: ActionMessage
    :members:

.. rst-class:: html-toggle

Message Parsers
---------------
.. autoclass:: MessageParser
    :members:

.. autoclass:: ActionParser
    :members:

.. rst-class:: html-toggle

Protocol Messages
-----------------
.. autoclass:: FeedbackMessage
    :members:

.. autoclass:: AddUserMessage
    :members:

.. autoclass:: RemoveUserMessage
    :members:

.. autoclass:: UpdateUserMessage
    :members:

.. autoclass:: AddKeyMessage
    :members:

.. autoclass:: RemoveKeyMessage
    :members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import re
import random

from bastio.mixin import Json, public
from bastio.ssh.crypto import RSAKey
from bastio.excepts import (BastioNetstringError, BastioEOFError,
        BastioMessageError, reraise)

@public
class Netstring(object):
    """A class to help parsing and composing Netstring formatted messages.

    This class takes a ``sock`` and an optional message length ``limit`` argument
    through its constructor for operating on data as it is received. The ``limit``
    parameter is in KiB, so passing ``limit=32`` means that the message length
    limit will be 32768 bytes. The socket passed must return a zero-sized string
    as an indication of EOF.

    There are two class methods, ``compose`` to compose Netstring formatted messages
    and ``parse`` to parse Netstring messages.
    """

    def __init__(self, sock, limit=32):
        self._sock = sock
        self._limit = limit * 1024

    def recv(self):
        """Receive a single Netstring message from the socket.

        :returns:
            The result of parsing a single Netstring message.
        """
        data_len = ''
        # Parse length part
        while True:
            char = self._sock.recv(1)
            if len(char) == 0:
                raise BastioEOFError("channel closed or EOF")
            data_len += char
            if data_len[-1] == ':':
                break
            elif not data_len[-1].isdigit():
                raise BastioNetstringError("non-digit character found in length part")
            elif int(data_len) > self._limit:
                raise BastioNetstringError("length part is bigger than the limit")
        try:
            data_len = int(data_len[:-1]) # Remove the extra ':'
        except ValueError:
            reraise(BastioNetstringError)
        data = self._sock.recv(data_len)
        if len(data) != data_len:
            raise BastioNetstringError("length specified does not match message length")
        if self._sock.recv(1) != ',':
            raise BastioNetstringError("message terminator is missing")
        return data

    @classmethod
    def compose(cls, data):
        """Compose a Netstring formatted message.

        :param data:
            The data to be wrapped in a Netstring message.
        :type data:
            str
        :returns:
            A Netstring formatted message for the data passed in.
        """
        return str(len(data)) + ":" + data + ","

    @classmethod
    def parse(cls, string):
        """Parse a Netstring message and return the result.

        :param string:
            The Netstring message to be parsed.
        :type string:
            str
        :returns:
            The result of parsing the Netstring message.
        """
        delim = string.find(':')
        if delim < 0:
            raise BastioNetstringError("unable to find length delimiter")
        elif delim == 0:
            raise BastioNetstringError("message length was not specified")

        try:
            length = int(string[:delim])
        except ValueError:
            reraise(BastioNetstringError)

        data = string[delim + 1:-1]
        if len(data) != length:
            raise BastioNetstringError("length specified does not match message length")
        if string[-1] != ',':
            raise BastioNetstringError("message terminator is missing")
        return data

@public
class ProtocolMessage(Json):
    """A protocol message base class."""

    def __init__(self, mid=None, **kwargs):
        super(ProtocolMessage, self).__init__()
        self.type = self.MessageType
        if not mid:
            # A new message construction, not a reply to a received message,
            # generate our own mid
            self.mid = self.__generate_mid()
        else:
            self.mid = mid
        self.parse(self, False)

    def reply(self, feedback, status):
        """Reply to a specific message with a feedback message that has
        the same MID.

        :param feedback:
            The feedback message string.
        :type feedback:
            str
        :param status:
            The status of the feedback message.
        :type status:
            int
        :returns:
            :class:`FeedbackMessage`
        """
        return FeedbackMessage(feedback, status, **self.__dict__)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check protocol message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        feedback object.

        :param obj:
            A JSON object containing the relevant fields for this protocol message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated protocol
            message object.
        """
        if 'mid' not in obj:
            raise BastioMessageError("message ID field is missing")
        if traverse:
            return cls(**obj.__dict__)

    @staticmethod
    def __generate_mid():
        return str(random.getrandbits(64))

@public
class FeedbackMessage(ProtocolMessage):
    """A protocol feedback message."""
    MessageType = "feedback"
    ERROR = 500
    WARNING = 400
    INFO = 300
    SUCCESS = 200
    STATUSES = [ERROR, WARNING, INFO, SUCCESS]

    def __init__(self, feedback, status, **kwargs):
        self.feedback = feedback
        self.status = status
        super(FeedbackMessage, self).__init__(**kwargs)
        self.parse(self, False)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check feedback message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        feedback object.

        :param obj:
            A JSON object containing the relevant fields for this feedback message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated feedback
            object.
        """
        if 'feedback' not in obj:
            raise BastioMessageError("feedback field is missing")
        if 'status' not in obj:
            raise BastioMessageError("status field is missing")
        if obj.status not in cls.STATUSES:
            raise BastioMessageError("status field is invalid")
        if traverse:
            return super(FeedbackMessage, cls).parse(obj)

@public
class ActionMessage(ProtocolMessage):
    """A protocol action message base class. Use this class as a base for all
    other action messages.
    """
    MessageType = 'action'

    def __init__(self, username, **kwargs):
        self.action = self.ActionType
        self.username = username
        super(ActionMessage, self).__init__(**kwargs)
        self.parse(self, False)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check action message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        action object.

        :param obj:
            A JSON object containing the relevant fields for this action message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated action
            object.
        """
        if 'username' not in obj:
            raise BastioMessageError("username field is missing")
        if not re.match("^([a-z_][a-z0-9_]{0,30})$", obj.username):
            raise BastioMessageError("username field is invalid")
        if traverse:
            return super(ActionMessage, cls).parse(obj)

@public
class AddUserMessage(ActionMessage):
    """An add-user action message."""
    ActionType = 'add-user'

    def __init__(self, sudo, **kwargs):
        self.sudo = sudo
        super(AddUserMessage, self).__init__(**kwargs)
        self.parse(self, False)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check add-user message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        action object.

        :param obj:
            A JSON object containing the relevant fields for this action message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated action
            object.
        """
        if 'sudo' not in obj:
            raise BastioMessageError("sudo field is missing")
        if traverse:
            return super(AddUserMessage, cls).parse(obj)

@public
class RemoveUserMessage(ActionMessage):
    """A remove-user action message."""
    ActionType = 'remove-user'

    def __init__(self, **kwargs):
        super(RemoveUserMessage, self).__init__(**kwargs)
        self.parse(self, False)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check remove-user message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        action object.

        :param obj:
            A JSON object containing the relevant fields for this action message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated action
            object.
        """
        if traverse:
            return super(RemoveUserMessage, cls).parse(obj)

@public
class UpdateUserMessage(ActionMessage):
    """A update-user action message."""
    ActionType = 'update-user'

    def __init__(self, sudo, **kwargs):
        self.sudo = sudo
        super(UpdateUserMessage, self).__init__(**kwargs)
        self.parse(self, False)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check update-user message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        action object.

        :param obj:
            A JSON object containing the relevant fields for this action message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated action
            object.
        """
        if 'sudo' not in obj:
            raise BastioMessageError("sudo field is missing")
        if traverse:
            return super(UpdateUserMessage, cls).parse(obj)

@public
class AddKeyMessage(ActionMessage):
    """A add-key action message."""
    ActionType = 'add-key'

    def __init__(self, public_key, **kwargs):
        self.public_key = public_key
        super(AddKeyMessage, self).__init__(**kwargs)
        self.parse(self, False)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check add-key message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        action object.

        :param obj:
            A JSON object containing the relevant fields for this action message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated action
            object.
        """
        if 'public_key' not in obj:
            raise BastioMessageError("public_key field is missing")
        if not RSAKey.validate_public_key(obj.public_key):
            raise BastioMessageError("public_key field is invalid")
        if traverse:
            return super(AddKeyMessage, cls).parse(obj)

@public
class RemoveKeyMessage(ActionMessage):
    """A remove-key action message."""
    ActionType = 'remove-key'

    def __init__(self, public_key, **kwargs):
        self.public_key = public_key
        super(RemoveKeyMessage, self).__init__(**kwargs)
        self.parse(self, False)

    @classmethod
    def parse(cls, obj, traverse=True):
        """Check remove-key message fields and validate them.

        Return a new object of type ``cls`` containing the validated
        action object.

        :param obj:
            A JSON object containing the relevant fields for this action message.
        :type obj:
            :class:`bastio.mixin.Json`
        :param traverse:
            Whether to traverse ``parse`` on all the classes in the hierarchy.
        :type traverse:
            bool
        :returns:
            A new object of type ``cls`` containing the validated action
            object.
        """
        if 'public_key' not in obj:
            raise BastioMessageError("public_key field is missing")
        if not RSAKey.validate_public_key(obj.public_key):
            raise BastioMessageError("public_key field is invalid")
        if traverse:
            return super(RemoveKeyMessage, cls).parse(obj)

@public
class ActionParser(object):
    """A protocol action message parser.

    Note that this class acts as a router for the actions that you can carry
    out, so in order to add support for a particular action you will have to
    edit the class dictionary ``SupportedActions`` to include the action type
    you wish to support.
    """
    MessageType = ActionMessage.MessageType
    SupportedActions = {
            AddUserMessage.ActionType: AddUserMessage,
            RemoveUserMessage.ActionType: RemoveUserMessage,
            UpdateUserMessage.ActionType: UpdateUserMessage,
            AddKeyMessage.ActionType: AddKeyMessage,
            RemoveKeyMessage.ActionType: RemoveKeyMessage,
            }

    @classmethod
    def parse(cls, obj):
        """Parse an action-type and return the relevant action object that
        represents the action-type of this message.

        :param obj:
            A JSON object for an action.
        :type obj:
            :class:`bastio.mixin.Json`
        :returns:
            A parsed and validated action message.
        """
        if 'action' not in obj:
            raise BastioMessageError("action field is missing")
        if obj.action not in cls.SupportedActions:
            raise BastioMessageError(
                    "action type `{}` is not supported".format(obj.action))
        return cls.SupportedActions[obj.action].parse(obj)

@public
class MessageParser(object):
    """A protocol message parser for JSON strings.

    Note that this class acts as a router for the type of messages you can
    handle, so in order to add support for another message you will have to
    edit the class dictionary ``SupportedMessages`` to include the message
    type you wish to support.
    """

    SupportedMessages = {
            FeedbackMessage.MessageType: FeedbackMessage,
            ActionParser.MessageType: ActionParser,
            }

    @classmethod
    def parse(cls, json_string):
        """Parse a JSON string and return the relevant object that represents
        the type of this message.

        :param json_string:
            The JSON string for a message.
        :type json_string:
            str
        :returns:
            An object that represents this message type.
        """
        try:
            obj = Json().from_json(json_string)
        except Exception:
            reraise(BastioMessageError)

        if 'type' not in obj:
            raise BastioMessageError("type field is missing")
        if obj.type not in cls.SupportedMessages:
            raise BastioMessageError(
                    "message type `{}` is not supported".format(obj.type))
        return cls.SupportedMessages[obj.type].parse(obj)




import unittest
import test_concurrency
import test_mixin
import test_configs
import test_ssh_crypto
import test_ssh_protocol
import test_ssh_api
import test_ssh_client

def __make_suite(tests):
    suite = unittest.TestSuite()
    for test in tests:
        suite.addTests(unittest.TestLoader().loadTestsFromTestCase(test))
    return suite

suite = unittest.TestSuite()
suite.addTests(__make_suite(test_concurrency.tests))
suite.addTests(__make_suite(test_mixin.tests))
suite.addTests(__make_suite(test_configs.tests))
suite.addTests(__make_suite(test_ssh_crypto.tests))
suite.addTests(__make_suite(test_ssh_protocol.tests))
suite.addTests(__make_suite(test_ssh_api.tests))
suite.addTests(__make_suite(test_ssh_client.tests))




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.test.test_ssh_protocol
:synopsis: Unit tests for the ssh.protocol module.
:author: Amr Ali <amr@databracket.com>
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import unittest
import StringIO

from bastio.mixin import Json
from bastio.excepts import BastioMessageError
from bastio.ssh.crypto import RSAKey
from bastio.ssh.protocol import (Netstring, MessageParser, ProtocolMessage,
        FeedbackMessage, ActionMessage, AddUserMessage, RemoveUserMessage,
        UpdateUserMessage, AddKeyMessage, RemoveKeyMessage, ActionParser)

class TestNetstring(unittest.TestCase):
    def test_netstring(self):
        data = 'hello world'
        enc = Netstring.compose(data)
        res = Netstring.parse(enc)
        self.assertEqual(data, res)

        buf = StringIO.StringIO(enc)
        buf.recv = buf.read
        net = Netstring(buf)
        res = net.recv()
        self.assertEqual(data, res)

class TestProtocolMessages(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pubkey = RSAKey.generate(1024).get_public_key()

    def test_message_add_user(self):
        obj = self._construct_action_msg(AddUserMessage)
        self._msg_parser_raises(obj)
        obj.sudo = False
        msg = MessageParser.parse(obj.to_json())
        self.assertIsInstance(msg, AddUserMessage)

    def test_message_remove_user(self):
        obj = self._construct_action_msg(RemoveUserMessage)
        msg = MessageParser.parse(obj.to_json())
        self.assertIsInstance(msg, RemoveUserMessage)

    def test_message_update_user(self):
        obj = self._construct_action_msg(UpdateUserMessage)
        self._msg_parser_raises(obj)
        obj.sudo = False
        msg = MessageParser.parse(obj.to_json())
        self.assertIsInstance(msg, UpdateUserMessage)

    def test_message_add_key(self):
        obj = self._construct_action_msg(AddKeyMessage)
        self._msg_parser_raises(obj)
        obj.public_key = self.pubkey
        msg = MessageParser.parse(obj.to_json())
        self.assertIsInstance(msg, AddKeyMessage)

    def test_message_remove_key(self):
        obj = self._construct_action_msg(RemoveKeyMessage)
        self._msg_parser_raises(obj)
        obj.public_key = self.pubkey
        msg = MessageParser.parse(obj.to_json())
        self.assertIsInstance(msg, RemoveKeyMessage)

    def test_message_feedback(self):
        obj = self._construct_protocol_msg()
        obj.type = FeedbackMessage.MessageType
        self._msg_parser_raises(obj)
        obj.feedback = "test"
        self._msg_parser_raises(obj)
        obj.status = 31337 # invalid status
        self._msg_parser_raises(obj)
        obj.status = FeedbackMessage.INFO
        msg = MessageParser.parse(obj.to_json())
        self.assertIsInstance(msg, FeedbackMessage)

    def test_message_reply(self):
        obj = self._construct_action_msg(RemoveUserMessage)
        msg = MessageParser.parse(obj.to_json())
        fb = msg.reply("test success", FeedbackMessage.SUCCESS)
        self.assertIsInstance(fb, FeedbackMessage)
        self.assertEqual(fb.feedback, "test success")
        self.assertEqual(fb.status, FeedbackMessage.SUCCESS)
        fb = msg.reply("test warning", FeedbackMessage.WARNING)
        self.assertIsInstance(fb, FeedbackMessage)
        self.assertEqual(fb.feedback, "test warning")
        self.assertEqual(fb.status, FeedbackMessage.WARNING)
        self.assertEqual(fb.mid, msg.mid)
        fb.to_json()

    def test_message_username(self):
        obj = self._construct_action_msg(AddUserMessage, username='@@@@23#$@FE__')
        self._msg_parser_raises(obj)
        obj.sudo = False
        self._msg_parser_raises(obj)

    def _msg_parser_raises(self, obj):
        with self.assertRaises(BastioMessageError):
            MessageParser.parse(obj.to_json())

    def _construct_protocol_msg(self):
        obj = Json()
        obj.mid = "234293472786"
        self._msg_parser_raises(obj)
        return obj

    def _construct_action_msg(self, action, username='d4de'):
        obj = self._construct_protocol_msg()
        obj.type = ActionMessage.MessageType
        self._msg_parser_raises(obj)
        obj.action = action.ActionType
        self._msg_parser_raises(obj)
        obj.username = username
        return obj

tests = [
        TestNetstring,
        TestProtocolMessages,
        ]




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.test.test_concurrency
:synopsis: Unit tests for the concurrency module.
:author: Amr Ali <amr@databracket.com>
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import time
import unittest

try:
    import queue
except ImportError:
    # Python 2.x
    import Queue as queue

from bastio.concurrency import ThreadPool, Task, Failure
from bastio.configs import GlobalConfigStore

class TestTask(unittest.TestCase):
    def test_task(self):
        task = Task(target=lambda x: x)
        self.assertIsInstance(task.id, long)
        self.assertEqual(task.target(True), True)
        self.assertEqual(task.success(True), True)
        self.assertEqual(task.failure(False), False)
        self.assertIsInstance(task.args, tuple)
        self.assertIsInstance(task.kwargs, dict)

class TestThreadPool(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        GlobalConfigStore().stacksize = 512
        cls.tp = ThreadPool(3)
        cls.tp_results = queue.Queue()

    @classmethod
    def tearDownClass(cls):
        cls.tp.remove_all_workers(10)

    def test_worker_task(self):
        t = Task(target=self.__test_target, success=self._test_cb)
        t.args = ("task1",)
        t = self.tp.run(t)
        retval = self.tp_results.get(timeout=5)
        self.assertEqual(retval, "task1test")

    def test_tp_stress(self):
        for x in range(5000):
            t = Task(target=self.__test_target)
            t.args = ("task" + str(x),)
            t = self.tp.run(t)

    def test_task_infinite(self):
        d = dict(counter = 0)
        t = Task(target=self.__test_counter, infinite=True)
        t.args = (d,)
        t = self.tp.run(t)
        time.sleep(0.01)
        t.stop()
        self.assertGreater(d['counter'], 2)

    def test_task_failure(self):
        t = Task(target=self.__test_failure, failure=self._catch_fail)
        self.tp.run(t)

    def _test_cb(self, retval):
        self.tp_results.put(retval)

    def _catch_fail(self, fail):
        with self.assertRaises(RuntimeError):
            raise fail.exception, fail.message, fail.traceback

    @staticmethod
    def __test_failure():
        raise RuntimeError

    @staticmethod
    def __test_target(arg):
        time.sleep(0.001) # 1ms
        return arg + "test"

    @staticmethod
    def __test_counter(kill_event, arg):
        arg['counter'] += 1

tests = [
        TestTask,
        TestThreadPool,
        ]




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.mixin
:synopsis: Mixins utilities used across the project.
:author: Amr Ali <amr@databracket.com>

.. rst-class:: html-toggle

Metaclasses
-----------
.. autoclass:: SingletonAbstractMeta

.. autoclass:: KindSingletonMeta

.. autoclass:: UniqueSingletonMeta

.. rst-class:: html-toggle

Data Structure Mixins
---------------------
.. autoclass:: Json
    :members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import json
from itertools import imap, ifilter
from ast import literal_eval
from collections import defaultdict

def public(obj):
    """A decorator to avoid retyping function/class names in __all__."""
    import sys
    _all = sys.modules[obj.__module__].__dict__.setdefault('__all__', [])
    if obj.__name__ not in _all:
        _all.append(obj.__name__)
    return obj

class SingletonAbstractMeta(type):
    """A Singleton pattern abstract metaclass."""

    def __init__(cls, name, bases, attrs):
        super(SingletonAbstractMeta, cls).__init__(name, bases, attrs)
        cls._instance = defaultdict(str)

    def __call__(cls, *args, **kwargs):
        key = cls._make_key(*args, **kwargs)
        if key not in cls._instance:
            cls._instance[key] = super(SingletonAbstractMeta, cls).__call__(
                    *args, **kwargs)
        return cls._instance[key]

    @classmethod
    def _make_key(cls, *args, **kwargs):
        """A function responsible for generating a key to check against and see
        if an instance was already constructed for that key.

        This is an abstract function, it is required to override it.
        """
        from bastio.excepts import BastioUnimplementedError
        raise BastioUnimplementedError("function _make_id is not implemented")

@public
class KindSingletonMeta(SingletonAbstractMeta):
    """A singleton pattern metaclass that uses the underlying class' name
    as the unique constraint.
    """

    @classmethod
    def _make_key(cls, *args, **kwargs):
        return cls.__class__

@public
class UniqueSingletonMeta(SingletonAbstractMeta):
    """A Singleton pattern metaclass that uses arguments passed to the
    underlying class constructor as the unique constraint.
    """

    @classmethod
    def _make_key(cls, *args, **kwargs):
        key = ''
        for arg in args:
            try:
                key += str(arg)
            except:
                continue
        for kwarg, value in kwargs.iteritems():
            try:
                key += str(kwarg) + str(value)
            except:
                continue
        return key

@public
class Json(object):
    """A mixin to give objects the ability to serialize and deserialize to/from
    JSON formatted string.
    """

    def to_json(self):
        """Serialize current object's members to a JSON formatted string.

        This function will only serialize members that don't start with ``_``
        or a ``callable``.

        :returns:
            A JSON formatted string.
        """
        res = {}
        for key, value in self.__dict__.iteritems():
            if key.startswith('_') or callable(value):
                continue

            if isinstance(value, self.__class__):
                res.update({ key: literal_eval(value.to_json()) })
            else:
                res.update({ key: value })
        return json.dumps(res)

    def from_json(self, json_string):
        """Deserialize a JSON formatted string and populate the current object
        with members of the name and value provided in the JSON string.

        :returns:
            ``self``
        """
        obj = json.loads(json_string)
        for key, value in obj.iteritems():
            if isinstance(value, dict):
                self.__dict__[key] = Json()
                self.__dict__[key].__dict__.update(value)
            else:
                self.__dict__[key] = value
        return self

    def __contains__(self, field):
        """Check whether a certain field exists in this object.

        :param field:
            The field name to be checked against __dict__.
        :type field:
            str
        :returns:
            bool
        """
        return field in imap(lambda x: x[0], ifilter(
                lambda x: not(x[0].startswith('_') or callable(x[1])),
                self.__dict__.iteritems()))




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.test.test_ssh_crypto
:synopsis: Unit tests for the ssh.crypto module.
:author: Amr Ali <amr@databracket.com>
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import os
import unittest

from bastio.ssh.crypto import RSAKey
from bastio.excepts import BastioCryptoError

class TestRSAKey(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.key = RSAKey.generate(1024)
        cls.key_file = 'TEST_PRIVATE_KEY'
        file(cls.key_file, 'wb').write(cls.key.get_private_key())

    @classmethod
    def tearDownClass(cls):
        os.unlink(cls.key_file)

    def test_key_generation(self):
        # Test invalid key size
        with self.assertRaises(BastioCryptoError):
            RSAKey.generate(1111)
        self.assertTrue(self.key.get_private_key())
        self.assertTrue(self.key.get_public_key())

    def test_key_validation(self):
        priv = self.key.get_private_key()
        self.assertTrue(RSAKey.validate_private_key(priv))
        self.assertTrue(RSAKey.validate_private_key_file('TEST_PRIVATE_KEY'))
        self.assertTrue(RSAKey.validate_public_key(self.key.get_public_key()))

    def test_key_loading(self):
        pubkey = self.key.get_public_key()
        self.assertIsInstance(RSAKey.from_public_key(pubkey), RSAKey)

tests = [
        TestRSAKey,
        ]




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.test.test_ssh_client
:synopsis: Unit tests for the ssh.client module.
:author: Amr Ali <amr@databracket.com>
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import unittest
import threading
import paramiko
import socket
import logging
import Queue as queue

from bastio.configs import GlobalConfigStore
from bastio.ssh.client import BackendConnector
from bastio.ssh.crypto import RSAKey
from bastio.ssh.protocol import (Netstring, MessageParser, AddUserMessage,
        FeedbackMessage)
from bastio.excepts import (BastioNetstringError, BastioMessageError,
        BastioEOFError)

# Disable paramiko logging
__plog = logging.getLogger('paramiko')
__plog.addHandler(logging.NullHandler())

class SSHSubsystem(paramiko.SubsystemHandler):
    def start_subsystem(self, name, transport, chan):
        while True:
            try:
                json_string = self._read_message(chan)
                msg = MessageParser.parse(json_string)
                reply = msg.reply("message received successfully",
                        FeedbackMessage.SUCCESS)
                self._write_message(chan, reply.to_json())
                return
            except BastioNetstringError as ex:
                msg = FeedbackMessage(ex.message, FeedbackMessage.ERROR)
                self._write_message(chan, msg.to_json())
                return
            except BastioMessageError as ex:
                msg = FeedbackMessage(ex.message, FeedbackMessage.ERROR)
                self._write_message(chan, msg.to_json())
                return
            except BastioEOFError:
                return

    @staticmethod
    def _read_message(chan):
        nets = Netstring(chan)
        return nets.recv()

    @staticmethod
    def _write_message(chan, data):
        nets = Netstring.compose(data)
        remaining = len(nets)
        while remaining > 0:
            n = chan.send(nets)
            if n <= 0:
                raise BastioEOFError("channel closed")
            remaining -= n

class SSHServer(paramiko.ServerInterface):
    def check_channel_request(self, kind, chanid):
        if kind == 'session':
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_auth_publickey(self, username, key):
        cfg = GlobalConfigStore()
        if (username == cfg.agent_username) and (key == cfg.agentkey):
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def get_allowed_auths(self, username):
        return 'publickey'

class TestBackendConnector(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Prepare server
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('127.0.0.1', 12345))
        sock.listen(10)
        cls.server_sock = sock
        th = threading.Thread(target=cls.__server)
        th.setDaemon(True)
        th.start()
        cls.server_thread = th
        cls.server_ready = threading.Event()
        cls.server_end = threading.Event()

        # Prepare BackendConnector
        cfg = GlobalConfigStore()
        cfg.agent_username = "test_agent"
        cfg.agentkey = RSAKey.generate(1024)
        cfg.backend_hostkey = RSAKey.generate(1024)
        cfg.host = "127.0.0.1"
        cfg.port = 12345
        cls.ingress = queue.Queue()
        cls.egress = queue.Queue()
        cls.connector = BackendConnector()
        endpoint = BackendConnector.EndPoint(ingress=cls.ingress,
                egress=cls.egress)
        cls.connector.register(endpoint)
        cls.connector.start()

    @classmethod
    def tearDownClass(cls):
        cls.connector.stop()
        cls.server_end.set()
        cls.server_thread.join()
        cls.server_sock.close()

    def test_backend_connector(self):
        self.server_ready.wait()
        msg = AddUserMessage(username="test_user", sudo=False)
        self.assertTrue(self.connector.is_active())
        self.egress.put(msg)

        reply = self.ingress.get(timeout=10)
        self.assertIsNotNone(reply)
        self.assertIsInstance(reply, FeedbackMessage)
        self.assertEqual(msg.mid, reply.mid)
        errmsg = "{} != {}: {}".format(reply.status, FeedbackMessage.SUCCESS,
                reply.feedback)
        self.assertEqual(reply.status, FeedbackMessage.SUCCESS, msg=errmsg)

    @classmethod
    def __server(cls):
        cfg = GlobalConfigStore()
        client, addr = cls.server_sock.accept()
        t = paramiko.Transport(client)
        t.add_server_key(cfg.backend_hostkey)
        t.set_subsystem_handler(cls.connector.Subsystem, SSHSubsystem)
        t.start_server(server=SSHServer())

        chan = t.accept(20)
        cls.server_ready.set()
        cls.server_end.wait()
        if chan:
            chan.close()
        t.close()

tests = [
        TestBackendConnector,
        ]




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.account
:synopsis: A module responsible for communicating account details.
:author: Amr Ali <amr@databracket.com>

.. autofunction:: download_backend_hostkey

.. autofunction:: upload_public_key
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import requests

from bastio import __version__
from bastio.mixin import public, Json
from bastio.ssh.crypto import RSAKey
from bastio.excepts import BastioAccountError, reraise

# Endpoints
__base_url = 'https://bastio.com/api/external/'
__download_hostkey_endpoint = __base_url + 'backend/host_key'
__upload_key_endpoint = __base_url + 'server/upload_key'

def __send_request(method, **kwargs):
    try:
        method = getattr(requests, method)
        client_version = "Bastio Agent v{}".format(__version__)
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers'].update({'User-Agent': client_version})
        return method(**kwargs)
    except requests.exceptions.SSLError:
        reraise(BastioAccountError, "SSL verification failed")
    except requests.exceptions.RequestException:
        reraise(BastioAccountError)
    except Exception as ex:
        reraise(BastioAccountError,
                "request unhandled exception occurred: " + ex.message)

@public
def download_backend_hostkey():
    """Get Bastio's backend SSH host key.

    :returns:
        :class:`bastio.ssh.crypto.RSAKey`
    :raises:
        :class:`bastio.excepts.BastioAccountError`
    """
    errmsg = "get backend host key failed: "

    response = __send_request('get', url=__download_hostkey_endpoint, verify=True)
    if response.status_code != requests.codes.okay: # 200
        raise BastioAccountError(errmsg + "unable to retrieve backend's host key")

    public_key = response.json()['payload']
    if not RSAKey.validate_public_key(public_key):
        raise BastioAccountError(errmsg + "invalid host key")
    return RSAKey.from_public_key(public_key)

@public
def upload_public_key(api_key, public_key, old_public_key=None):
    """Upload agent's public key to Bastio on the account specified by ``api_key``.

    This action will create a new server when ``old_public_key`` is not specified
    and it is the first time we see this ``public_key``. However the public key
    we store that identifies this server will be replaced IFF we already have
    ``old_public_key`` in our records and ``public_key`` does not exist in our
    database.

    :param api_key:
        The API key for the Bastio account.
    :type api_key:
        str
    :param public_key:
        The agent's public key to be uploaded to Bastio's servers.
    :type public_key:
        The string output of :func:`bastio.ssh.crypto.RSAKey.get_public_key`.
    :param old_public_key:
        The agent's old public key to be replaced by a new one.
    :type old_public_key:
        The string output of :func:`bastio.ssh.crypto.RSAKey.get_public_key`.
    :raises:
        :class:`bastio.excepts.BastioAccountError`
    """
    errmsg = "upload public key failed: "

    if not old_public_key:
        old_public_key = ''

    if old_public_key and not RSAKey.validate_public_key(old_public_key):
        raise BastioAccountError(errmsg + "invalid old public key")

    if not RSAKey.validate_public_key(public_key):
        raise BastioAccountError(errmsg + "invalid new public key")

    payload = Json()
    payload.api_key = api_key
    payload.public_key = public_key
    payload.old_public_key = old_public_key

    headers = {'Content-type': 'application/json'}
    response = __send_request('post', url=__upload_key_endpoint, verify=True,
            data=payload.to_json(), headers=headers)
    if response.status_code == requests.codes.bad: # 400
        raise BastioAccountError(errmsg + "missing or invalid field")
    elif response.status_code == requests.codes.forbidden: # 403
        raise BastioAccountError(errmsg + "not authorized or invalid API key")
    elif response.status_code == requests.codes.okay: # 200
        return

    # An unexpected response status code
    raise BastioAccountError(
            errmsg + "unexpected response status code ({})".format(
                response.status_code))




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.cli
:synopsis: A module responsible for the CLI of the agent.
:author: Amr Ali <amr@databracket.com>

.. autoclass:: CommandLine
    :members:

.. autofunction:: bastio_main
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import os
import sys
import signal
import argparse

from bastio import __version__
from bastio.log import Logger
from bastio.mixin import public
from bastio.configs import GlobalConfigStore
from bastio.concurrency import GlobalThreadPool
from bastio.account import upload_public_key, download_backend_hostkey
from bastio.ssh.client import BackendConnector
from bastio.ssh.api import Processor
from bastio.ssh.crypto import RSAKey
from bastio.excepts import BastioConfigError

def __sig_handler(sig, frame):
    logger = Logger()
    logger.critical("signal received, shutting down")
    cfg = GlobalConfigStore()
    cfg.connector.stop()
    cfg.processor.stop()
    cfg.threadpool.remove_all_workers(3)

def _check_file_readability(filename):
    # Return a tuple of two status indicators, the first is to indicate that the
    # file exists and the second is an indication of file's readability.
    if not(filename and os.path.exists(filename)):
        return (False, False)
    try:
        with open(filename, 'rb') as fd:
            return (True, True)
    except Exception:
        return (True, False)

def _die(msg, success=False):
    if success:
        sys.stdout.write("{}\n".format(msg))
        sys.exit(0)
    else:
        sys.stderr.write("error: {}\n".format(msg))
        sys.exit(1)

@public
class CommandLine(object):
    """Parse and logically validate command line argments."""
    _description = "Bastio agent responsible for provisioning\n" \
            "system accounts."
    _epilog = "Report Bastio's bugs to support@bastio.com"

    def __init__(self):
        parser = argparse.ArgumentParser(description=self._description,
                formatter_class=argparse.RawDescriptionHelpFormatter,
                epilog=self._epilog)

        # Configure the command line parser

        # Miscellaneous arguments
        parser.add_argument('-c', '--config',
                help='configuration file path')
        parser.add_argument('--debug', action='store_true',
                help='run the agent in debugging mode where the logging will output to STDOUT')
        parser.add_argument('--version', action='version',
                version='%(prog)s ' + __version__,
                help='output version information and exit')
        parser.add_argument('-k', '--agent-key',
                help='path to the agent private key')

        # A group for commands that require start action details
        start_group = argparse.ArgumentParser(add_help=False)
        start_group.add_argument('-H', '--host', default='backend.bastio.com',
                help='host name of the Bastio backend (default: %(default)s)')
        start_group.add_argument('-p', '--port', type=int, default=2357,
                help='port of the backend to connect to (default: %(default)s)')
        start_group.add_argument('-m', '--min-threads', type=int, default=3,
                help='the minimum number of threads the thread pool must have (default: %(default)s)')
        start_group.add_argument('-s', '--stack-size', type=int, default=512,
                help='the stack size of each thread in KiB (default: %(default)sKiB)')

        # A group for commands that require account details
        api_group = argparse.ArgumentParser(add_help=False)
        api_group.add_argument('--api-key',
                help='Bastio API key')
        api_group.add_argument('-n', '--new-agent-key',
                help=('path to the new agent key to replace the old one specified by -k'
                    ' (this argument only make sense with `upload-key`)'))

        # A group for commands that require key details
        key_group = argparse.ArgumentParser(add_help=False)
        key_group.add_argument('--bits', type=int, default=2048,
                help=('number of bits to generate for the private key'
                    ' (default: %(default)s-bits, this argument only make sense with `generate-key`)'))

        # Command parsers
        sp = parser.add_subparsers(help='available commands', dest='command')
        sp.add_parser('generate-key', parents=[key_group],
                description=('Generate a new RSA private key and save it in the file '
                    'specified on the command line or in the configuration file.'),
                help='generate a new RSA private key for the agent')
        sp.add_parser('upload-key', parents=[api_group],
                description=('Extract the public key from the private key file '
                    'and upload it to Bastio server'),
                help='upload this agent public key')
        sp.add_parser('start', parents=[start_group, api_group],
                description='Start the agent in the foreground',
                help='start the agent')
        self.parser = parser

    def parse(self):
        """Parse and validate arguments from the command line and set
        global configurations.
        """
        self.args = self.parser.parse_args()
        cfg = GlobalConfigStore()
        cfg.prog = self.parser.prog
        cfg.debug = self.args.debug

        # Check of configuration file is available to us
        conf_avail = False
        if self.args.config:
            try:
                cfg.load(self.args.config)
                conf_avail = True
            except BastioConfigError as ex:
                self.parser.error(ex.message)

        # Check and validate agent's key if we are about to upload the key to
        # Bastio's servers or we are about to start the agent
        if self.args.command in ('upload-key', 'start'):
            # Get agent key file path from configuration file (if available)
            # or from the command line argument
            try:
                if conf_avail:
                    cfg.apikey = cfg.apikey if cfg.get_apikey else \
                            self.args.api_key
                    cfg.agentkey = cfg.agentkey if cfg.get_agentkey else \
                            self.args.agent_key
                else:
                    cfg.apikey = self.args.api_key
                    cfg.agentkey = self.args.agent_key
            except BastioConfigError as ex:
                _die(ex.message)
            # Check agent's key file readability and validate it
            res = _check_file_readability(cfg.agentkey)
            if not res[0]:
                self.parser.error('agent key file `{}` does not exist'.format(
                    cfg.agentkey))
            if not res[1]:
                self.parser.error(('permission to read the agent key file `{}` '
                    'is denied').format(cfg.agentkey))
            res = RSAKey.validate_private_key_file(cfg.agentkey)
            if not res:
                self.parser.error('agent key file `{}` is invalid'.format(
                    cfg.agentkey))

        # Parse and validate commands and their arguments
        if self.args.command == 'generate-key':
            try:
                if conf_avail:
                    cfg.agentkey = cfg.agentkey if cfg.get_agentkey else \
                            self.args.agent_key
                else:
                    cfg.agentkey = self.args.agent_key
                cfg.bits = self.args.bits
            except BastioConfigError as ex:
                _die(ex.message)
        elif self.args.command == 'upload-key':
            try:
                # Check new key file's readability and validate it if provided
                new_key = self.args.new_agent_key
                if new_key:
                    res = _check_file_readability(new_key)
                    if not res[0]:
                        self.parser.error(
                                'new agent key file `{}` does not exist'.format(
                                    new_key))
                    if not res[1]:
                        self.parser.error((
                                'permission to read the new agent key file `{}` '
                                'is denied').format(new_key))
                    res = RSAKey.validate_private_key_file(new_key)
                    if not res:
                        self.parser.error(
                                'new agent key file `{}` is invalid'.format(
                                    new_key))
                    cfg.new_agentkey = new_key
            except BastioConfigError as ex:
                _die(ex.message)
        elif self.args.command == 'start':
            try:
                if conf_avail:
                    cfg.host = cfg.host if cfg.get_host else self.args.host
                    cfg.port = cfg.port if cfg.getint_port else self.args.port
                    cfg.stacksize = cfg.stacksize if cfg.getint_stacksize else \
                            self.args.stack_size
                    cfg.minthreads = cfg.minthreads if cfg.getint_minthreads else \
                            self.args.min_threads
                else:
                    cfg.host = self.args.host
                    cfg.port = self.args.port
                    cfg.stacksize = self.args.stack_size
                    cfg.minthreads = self.args.min_threads
            except BastioConfigError as ex:
                _die(ex.message)
        else:
            # NOTE: This execution branch is blocked by argparse
            # so it is here only to account for extremely unlikely cases
            _die("unsupported command `{}`".format(self.args.command))
        return self.args.command

@public
def bastio_main():
    """Main application entry point."""
    signal.signal(signal.SIGINT, __sig_handler)
    signal.signal(signal.SIGTERM, __sig_handler)
    cfg = GlobalConfigStore()

    # Parse command line arguments
    cmd = CommandLine()
    command = cmd.parse()

    if command == 'generate-key':
        try:
            key = RSAKey.generate(cfg.bits)
            key.write_private_key_file(cfg.agentkey)
        except Exception as ex:
            _die(ex.message)
        _die("generated {}-bit key successfully".format(cfg.bits), True)
    elif command == 'upload-key':
        try:
            agentkey = RSAKey.from_private_key_file(cfg.agentkey)
        except Exception as ex:
            _die(ex.message)

        try:
            # Replace old key
            if cfg.new_agentkey:
                new_agentkey = RSAKey.from_private_key_file(cfg.new_agentkey)
                upload_public_key(cfg.apikey, new_agentkey.get_public_key(),
                        agentkey.get_public_key())
            else:
                upload_public_key(cfg.apikey, agentkey.get_public_key())
        except Exception as ex:
            _die(ex.message)
        _die("uploaded public key successfully", True)
    elif command == 'start':
        try:
            cfg.agent_username = cfg.apikey
            cfg.agentkey = RSAKey.from_private_key_file(cfg.agentkey)
            cfg.backend_hostkey = download_backend_hostkey()
        except Exception as ex:
            _die(ex.message)

    ### All that is below is part of the ``start`` command

    # Set logging based on debug status
    if cfg.debug:
        Logger().enable_stream()
    else:
        Logger().enable_syslog()

    cfg.threadpool = GlobalThreadPool(cfg.minthreads)
    cfg.processor = Processor()
    cfg.connector = BackendConnector()
    cfg.connector.register(cfg.processor.endpoint())
    cfg.connector.start()
    signal.pause()




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.ssh.api
:synopsis: A module responsible for the API between the backend and the agent.
:author: Amr Ali <amr@databracket.com>

.. autoclass:: Processor
    :members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import os
import pwd
import threading
import subprocess
import collections
import Queue as queue

from bastio.log import Logger
from bastio.mixin import KindSingletonMeta, public
from bastio.configs import GlobalConfigStore
from bastio.concurrency import GlobalThreadPool, Task
from bastio.ssh.client import BackendConnector
from bastio.ssh.protocol import (FeedbackMessage, AddUserMessage,
        RemoveUserMessage, UpdateUserMessage, AddKeyMessage, RemoveKeyMessage)

@public
class Processor(object):
    """A class to handle action messages coming from the backend and send back
    a feedback to indicate success or failure of the action requested. This class
    is a kind-singleton which means you cannot instantiate more than one copy per
    application life time.
    """
    __metaclass__ = KindSingletonMeta

    def __init__(self):
        self._tp = GlobalThreadPool()
        self._logger = Logger()
        self._ingress = queue.Queue()
        self._egress = queue.Queue()
        # TODO: Put the following in a configuration file.
        self._home_dir = '/home'
        self._user_dir = os.path.join(self._home_dir, '{username}')
        self._ssh_dir = os.path.join(self._user_dir, '.ssh')
        self._authkeys = os.path.join(self._ssh_dir, 'authorized_keys')
        # Start the action handler
        t = Task(target=self.__action_handler, infinite=True)
        t.failure = self.__catch_fail
        self._action_handler_task = self._tp.run(t)

    def endpoint(self):
        """Return an ingress and an egress points to communicate with this
        processor.

        :returns:
            :class:`bastio.ssh.client.BackendConnector.EndPoint`
        """
        return BackendConnector.EndPoint(ingress=self._ingress, egress=self._egress)

    def process(self, message):
        """Process a message and return a feedback.

        :param message:
            A message to be processed.
        :type message:
            A subclass of :class:`bastio.ssh.protocol.ActionMessage`
        :returns:
            :class:`bastio.ssh.protocol.FeedbackMessage`
        """
        if isinstance(message, AddUserMessage):
            # Add a user if one doesn't exist
            feedback = self._add_user(message)
        elif isinstance(message, RemoveUserMessage):
            # Remove a user if one exists
            feedback = self._remove_user(message)
        elif isinstance(message, UpdateUserMessage):
            # Update a user either to give it root access or to demote it
            feedback = self._update_user(message)
        elif isinstance(message, AddKeyMessage):
            # Add public key to the user's authorized_keys file
            feedback = self._add_key(message)
        elif isinstance(message, RemoveKeyMessage):
            # Remove public key from the user's authorized_keys file
            feedback = self._remove_key(message)
        else:
            # NOTE: This execution branch must never be reached,
            # do not take this lightly if it happens.
            feedback = message.reply(
                    ("internal error: agent does not know how to handle messages"
                        " of type `{type}`").format(type=message.type),
                    FeedbackMessage.ERROR)
        return feedback

    def stop(self):
        """Signal the action handler to stop."""
        self._action_handler_task.stop()

    def __action_handler(self, kill_ev):
        self._logger.warning("action handler started")
        while not kill_ev.is_set():
            message = self._get_ingress(timeout=3)
            if message:
                feedback = self.process(message)
                self._put_egress(feedback)

    def __catch_fail(self, failure):
        try:
            raise failure.exception, failure.message, failure.traceback
        except Exception:
            self._logger.critical("unexpected error occurred in the action handler",
                    exc_info=True)

    def _get_ingress(self, timeout):
        try:
            return self._ingress.get(timeout=timeout)
        except queue.Empty:
            return None

    def _put_egress(self, item):
        self._egress.put(item)

###
###  BEGIN COMMAND METHODS
###

    def _chk_user(self, message, status=FeedbackMessage.ERROR, should_exist=False):
        # Check if a user exists
        user_exist = os.path.exists(self._user_dir.format(
            username=message.username))
        try:
            pwd.getpwnam(message.username)
        except KeyError:
            user_exist = False

        if user_exist:
            reply_msg = "{username} already exists".format(username=message.username)
            if should_exist: # user exists and should
                return False
            else: # user exists but shouldn't
                feedback = message.reply(reply_msg, status)
        else:
            reply_msg = "{username} does not exist".format(username=message.username)
            if should_exist: # user doesn't exist but should
                feedback = message.reply(reply_msg, status)
            else: # user doesn't exist and shouldn't
                return False
        return feedback

    def _chk_key(self, message):
        # Check if a public key exists
        try:
            with open(self._authkeys.format(username=message.username), 'rb') as fd:
                auth_data = fd.read()
        except Exception:
            return False
        return message.public_key in auth_data

    def _create_ssh(self, message):
        # Make sure that .ssh exists and has the right permissions
        try:
            os.mkdir(self._ssh_dir.format(username=message.username), 0700)
        except OSError:
            pass # Directory already exists (or perm denied... very unlikely)

        # Touch .ssh/authorized_keys file
        auth_file = self._authkeys.format(username=message.username)
        try:
            with open(auth_file, 'ab') as fd:
                pass # We just want to create the file if it doesn't exist
        except IOError:
            # We can't do anything about it here, it will be handled by other messages
            pass

        # Make sure that .ssh/authorized_keys file has the right permissions
        try:
            os.chmod(auth_file, 0600)
        except OSError:
            # We can't do anything about it here, offload it to future messages
            pass

        # Chown .ssh/authorized_keys to the user
        try:
            pw_struct = pwd.getpwnam(message.username)
            os.chown(self._ssh_dir.format(username=message.username),
                    pw_struct.pw_uid, pw_struct.pw_gid)
            os.chown(auth_file, pw_struct.pw_uid, pw_struct.pw_gid)
        except KeyError:
            # username not found from getpwnam
            pass
        except OSError:
            # chown failed
            pass

    def _add_user(self, message):
        # Add user
        if message.sudo:
            add_command = 'useradd -mU -G sudo {username}'
        else:
            add_command = 'useradd -mU {username}'
        add_command = add_command.format(username=message.username)

        # Check if a user exists
        feedback = self._chk_user(message, FeedbackMessage.INFO, False)
        if feedback:
            self._create_ssh(message)
            return feedback

        # Create the user
        _, stderr = self._run_command(add_command)
        if stderr:
            feedback = message.reply(stderr, FeedbackMessage.ERROR)
            return feedback

        # Clear out user's password
        _, stderr = self._run_command("passwd -d {username}".format(
            username=message.username))
        if stderr:
            feedback = message.reply(stderr, FeedbackMessage.ERROR)
            return feedback

        self._create_ssh(message)
        feedback = message.reply("{username} was created successfully".format(
            username=message.username), FeedbackMessage.SUCCESS)
        return feedback

    def _remove_user(self, message):
        # Remove user
        rm_command = 'userdel -r {username}'.format(username=message.username)

        # Check if a user exists
        feedback = self._chk_user(message, FeedbackMessage.INFO, True)
        if feedback:
            return feedback

        # Try to remove the user
        _, stderr = self._run_command(rm_command)
        if stderr:
            feedback = message.reply(stderr, FeedbackMessage.ERROR)
        else:
            feedback = message.reply(
                    "{username} was removed successfully".format(
                        username=message.username), FeedbackMessage.SUCCESS)
        return feedback

    def _update_user(self, message):
        # Update user
        flag = '-a' if message.sudo else '-d'
        update_command = 'gpasswd {flag} {username} sudo'.format(flag=flag,
                username=message.username)

        # Check if a user exists
        feedback = self._chk_user(message, FeedbackMessage.ERROR, True)
        if feedback:
            return feedback

        # Update a user either to give it root access or to demote it
        _, stderr = self._run_command(update_command)
        if stderr:
            feedback = message.reply(stderr, FeedbackMessage.ERROR)
        else:
            if message.sudo:
                fb_str = '{username} was added to the sudo group successfully'
            else:
                fb_str = '{username} was removed from the sudo group successfully'
            feedback = message.reply(fb_str.format(username=message.username),
                    FeedbackMessage.SUCCESS)
        return feedback

    def _add_key(self, message):
        # Add public key
        pubkey = message.public_key
        username = message.username

        # Check if a user exists
        feedback = self._chk_user(message, FeedbackMessage.ERROR, True)
        if feedback:
            return feedback

        # Check if public key already exists
        if self._chk_key(message):
            feedback = message.reply(
                    "public key `{pub_key}` for {username} already exists".format(
                        pub_key=pubkey, username=username),
                    FeedbackMessage.INFO)
            return feedback

        # Try to add the public key to the user's authorized_keys file
        auth_file = self._authkeys.format(username=username)
        try:
            with open(auth_file, 'ab') as fd:
                fd.write(pubkey + '\n')
            feedback = message.reply(
                    "added public key to {username} successfully".format(
                        username=username), FeedbackMessage.SUCCESS)
        except IOError as ex:
            feedback = message.reply(ex.strerror, FeedbackMessage.ERROR)
        except Exception as ex:
            feedback = message.reply(ex.message, FeedbackMessage.ERROR)
        return feedback

    def _remove_key(self, message):
        # Remove public key
        pubkey = message.public_key
        username = message.username

        # Check if a user exists
        feedback = self._chk_user(message, FeedbackMessage.ERROR, True)
        if feedback:
            return feedback

        # Check if public key does not exist
        if not self._chk_key(message):
            feedback = message.reply(
                    "public key for {username} does not exist".format(
                        username=username), FeedbackMessage.INFO)
            return feedback

        # Try to remove the public key from the user's authorized_keys file
        auth_data = []
        auth_file = self._authkeys.format(username=username)
        try:
            with open(auth_file, 'rb') as fd:
                for line in fd.readlines():
                    if pubkey in line:
                        continue
                    auth_data.append(line)
            with open(auth_file, 'wb') as fd:
                # TODO: A race condition is possible here where the file could
                # be written to before we write to it and therefore overriding
                # the changes made to it by some other application. Find a fix
                # for it. This is quite unlikely in this particular case so
                # don't sweat it.
                fd.writelines(auth_data)
            feedback = message.reply(
                    "removed public key from {username} successfully".format(
                        username=username), FeedbackMessage.SUCCESS)
        except IOError as ex:
            feedback = message.reply(ex.strerror, FeedbackMessage.ERROR)
        except Exception as ex:
            feedback = message.reply(ex.message, FeedbackMessage.ERROR)
        return feedback

    @staticmethod
    def _run_command(command, input_data=None):
        try:
            po = subprocess.Popen(args=command, shell=True, stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = po.communicate(input_data)
        except OSError as ex:
            stderr = ex.strerror
        except ValueError as ex:
            stderr = ex.message
        return stdout, stderr

###
###  END COMMAND METHODS
###




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.concurrency
:synopsis: Concurrency utilities.
:author: Amr Ali <amr@databracket.com>

.. autoclass:: Failure
    :members:

.. autoclass:: Task
    :members:

.. autoclass:: ThreadPool
    :members:

.. autoclass:: GlobalThreadPool
    :inherited-members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import sys
import random
import threading
import Queue as queue

from bastio.mixin import KindSingletonMeta, public
from bastio.excepts import BastioTaskError
from bastio.configs import GlobalConfigStore
from bastio.log import Logger

@public
class Failure(object):
    """A class to wrap an exception info to pass to the failure callback registered
    on a task to handle it.

    This class must only be instantiated from under an ``except`` clause.
    """

    def __init__(self):
        ex, msg, tb = sys.exc_info()
        self._exception = ex
        self._message = msg
        self._traceback = tb

    @property
    def exception(self):
        return self._exception

    @property
    def message(self):
        return self._message

    @property
    def traceback(self):
        return self._traceback

@public
class Task(object):
    """A class to describe a task to be used by the thread pool.

    An infinite or a blocking task will have to take a kill :class:`threading.Event`
    as the first argument to the function to be able to end operations gracefully
    in case a stop event was triggered.

    This class takes two callbacks as keyword arguments to handle the cases where
    a task succeeds or fails, called ``success`` and ``failure`` respectively.

    Calling the :func:`Task.stop` method will signal the worker in the thread pool
    to stop execution and exit gracefully.
    """

    def __init__(self, target, success=lambda x: x, failure=lambda x: x, infinite=False,
            *args, **kwargs):
        self._id = random.getrandbits(128)
        self._kill_ev = threading.Event()
        self._infinite = infinite
        self.target = target
        self.success = success
        self.failure = failure
        self.args = args
        self.kwargs = kwargs

    def stop(self):
        """Signal this task to stop as soon as possible."""
        self._infinite = False
        self._kill_ev.set()

    @property
    def id(self):
        return self._id

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        if not callable(value):
            raise BastioTaskError("field target need to be a callable")
        self._target = value

    @property
    def success(self):
        return self._success

    @success.setter
    def success(self, value):
        if not callable(value):
            raise BastioTaskError("field success need to be a callable")
        self._success = value

    @property
    def failure(self):
        return self._failure

    @failure.setter
    def failure(self, value):
        if not callable(value):
            raise BastioTaskError("field failure need to be a callable")
        self._failure = value

    @property
    def infinite(self):
        return self._infinite

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, value):
        if not isinstance(value, (tuple, list)):
            raise BastioTaskError("field args need to be either a list or a tuple")
        value = tuple(value)
        if self.infinite:
            self._args = (self._kill_ev,) + value
        else:
            self._args = value

    @property
    def kwargs(self):
        return self._kwargs

    @kwargs.setter
    def kwargs(self, value):
        if not isinstance(value, dict):
            raise BastioTaskError("field kwargs need to be a dict")
        self._kwargs = value

@public
class ThreadPool(object):
    """An adaptive thread pool.

    This thread pool adapts to the threads consumption rate by making sure that
    the available number of workers to process tasks is always above the minimum
    number of workers that should be always available.
    """

    ThreadFactory = threading.Thread

    def __init__(self, min_workers = 2):
        stacksize = GlobalConfigStore().stacksize
        threading.stack_size(stacksize * 1024)
        self._logger = Logger()
        self._tasks = queue.Queue()
        self._running_tasks = []
        self._min_workers = min_workers + 1 # for the monitoring thread
        self._workers = 0
        self._avail_workers = 0
        self._countlck = threading.Lock()
        self._task_added = threading.Event()
        self._killev = threading.Event()
        self._all_died = threading.Event()
        self.add_worker(self._min_workers)
        mt = Task(target=self.__volume_monitor, infinite=True)
        self.run(mt)

    def run(self, task):
        """Start a task.

        :param task:
            A task to be executed by a worker.
        :type task:
            :class:`Task`
        :returns:
            The task that was passed to this method.
        """
        self._task_added.set()
        self._tasks.put(task)
        return task

    def add_worker(self, num=1):
        """Add worker(s) to the thread pool.

        :param num:
            The number of workers to add to the pool.
        :type num:
            int
        """
        for x in range(int(num)):
            t = self.ThreadFactory(target=self.__worker)
            t.setDaemon(True)
            t.start()

    def remove_worker(self, num=1):
        """Remove worker(s) from the thread pool.

        :param num:
            The number of workers to remove from the pool.
        :type num:
            int
        """
        for x in range(int(num)):
            self._tasks.put("exit")

    def remove_all_workers(self, wait=None):
        """Remove all workers from the pool.

        Remove all active workers from the pool and wait ``wait`` seconds
        until last worker ends, or wait forever if ``wait`` is None. This
        action will also signal all running tasks to stop as soon as possible.

        :param wait:
            Number of seconds to wait or None to wait forever.
        :type wait:
            float
        """
        self._killev.set()
        self.remove_worker(self._workers)
        self._task_added.set()
        for task in self._running_tasks:
            task.stop()
        self._all_died.wait(wait)
        self._killev.clear()

    def __volume_monitor(self, kill_ev):
        while not kill_ev.is_set():
            with self._countlck:
                if self._workers < self._min_workers:
                    self.add_worker(self._min_workers - self._workers)
                if self._avail_workers < self._min_workers:
                    self.add_worker(round(abs(self._workers - self._avail_workers) / 2.0))
            self._task_added.wait(5.0)
            self._task_added.clear()

    def __worker(self):
        with self._countlck:
            self._workers += 1
            self._avail_workers += 1
            self._all_died.clear()
        while not self._killev.is_set(): # Main thread body
            try:
                task = self._tasks.get(timeout=1.0)
            except queue.Empty:
                # Waited for too long
                break
            if task == 'exit': # "exit" is a sentinel task to kill the worker
                break

            with self._countlck:
                self._avail_workers -= 1

            # Execute target function here
            self._running_tasks.append(task)
            try:
                ret = task.target(*task.args, **task.kwargs)
                if task.success:
                    task.success(ret)
            except Exception as ex:
                if task.failure:
                    try:
                        task.failure(Failure())
                    except Exception:
                        msg = 'failure callback raised an error on task ({})'.format(task.id)
                        self._logger.critical(msg, exc_info=True)
                else:
                    msg = "unhandled error occurred on task ({}): {}".format(
                            task.id, ex.message)
                    self._logger.critical(msg, exc_info=True)
            self._running_tasks.remove(task)

            if task.infinite:
                self._tasks.put(task)
            with self._countlck:
                self._avail_workers += 1
        with self._countlck:
            self._workers -= 1
            self._avail_workers -= 1
            if not self._workers:
                self._all_died.set()

@public
class GlobalThreadPool(ThreadPool):
    """A singleton of :class:`bastio.concurrency.ThreadPool`."""
    __metaclass__ = KindSingletonMeta




# Copyright 2013 Databracket LLC.
# See LICENSE file for details.

"""
.. rst-class:: html-toggle

CLI Management
--------------
.. automodule:: bastio.cli

.. rst-class:: html-toggle

SSH And Communication Management
--------------------------------
.. automodule:: bastio.ssh

.. rst-class:: html-toggle

Account Operations
------------------
.. automodule:: bastio.account

.. rst-class:: html-toggle

Configuration Utilities
-----------------------
.. automodule:: bastio.configs

.. rst-class:: html-toggle

Concurrency Utilities
---------------------
.. automodule:: bastio.concurrency

.. rst-class:: html-toggle

Logging Facility
----------------
.. automodule:: bastio.log

.. rst-class:: html-toggle

Mixins And Language Helpers
---------------------------
.. automodule:: bastio.mixin

.. rst-class:: html-toggle

Version Information
-------------------
.. automodule:: bastio.version

.. rst-class:: html-toggle

Exceptions
----------
.. automodule:: bastio.excepts
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

from .version import __version__

def load_test_suite():
    from bastio.test import suite
    return suite




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.ssh.client
:synopsis: A module for SSH client implementations.
:author: Amr Ali <amr@databracket.com>

.. autoclass:: BackendConnector
    :members:
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import time
import socket
import paramiko
import collections
import Queue as queue

from bastio import __version__
from bastio.log import Logger
from bastio.mixin import KindSingletonMeta, public
from bastio.configs import GlobalConfigStore
from bastio.concurrency import GlobalThreadPool, Task
from bastio.ssh.protocol import Netstring, MessageParser, ProtocolMessage
from bastio.excepts import (BastioBackendError, BastioEOFError,
        BastioNetstringError, BastioMessageError, reraise)

# Set paramiko client ID
paramiko.Transport._CLIENT_ID = "bastio-{}".format(__version__)

@public
class BackendConnector(object):
    """A singleton to establish and maintain a secure connection with the backend
    over a specific subsystem channel. This connector supports registering of
    endpoints where processors can register their endpoint to communicate with
    the backend. It is guaranteed that the messages will be delivered ASAP but
    the actual ETA is chaotic.
    """
    __metaclass__ = KindSingletonMeta
    EndPoint = collections.namedtuple("EndPoint", "ingress egress")
    Subsystem = 'bastio-agent'

    def __init__(self):
        cfg = GlobalConfigStore()
        self._tp = GlobalThreadPool()
        self._username = cfg.agent_username
        self._agent_key = cfg.agentkey
        self._backend_addr = (cfg.host, cfg.port)
        self._backend_hostkey = cfg.backend_hostkey
        self._logger = Logger()
        self._endpoints = []
        self._tx = queue.Queue()
        self._conn_handler_task = None
        self._connected = False
        self._running = False
        self._client = None
        self._chan = None

    def start(self):
        """Start the connection handler thread."""
        if not self._running:
            self._running = True
            t = Task(target=self.__conn_handler, infinite=True)
            t.failure = self._catch_fail
            self._conn_handler_task = self._tp.run(t)

    def stop(self):
        """Stop the connection handler thread."""
        if self._running:
            self._running = False
            self.close()
            self._conn_handler_task.stop()

    def register(self, endpoint):
        """Register an endpoint to this connector to so that it can communicate
        with the backend. The endpoint is a tuple of one ingress queue as first
        argument and egress as the second argument.

        :param endpoint:
            A tuple of two queues; ingress and egress.
        :type endpoint:
            :class:`BackendConnector.EndPoint`
        """
        self._endpoints.append(endpoint)

    def is_active(self):
        """Check whether the transport is still active."""
        if self._client:
            t = self._client.get_transport()
            if t:
                return t.is_active()
        return False

    def close(self):
        """Close open channels and transport."""
        if self._chan:
            self._chan.close()
        if self._client:
            self._client.close()
        self._connected = False
        self._logger.critical("connection lost with the backend")

    def __conn_handler(self, kill_ev):
        self._logger.warning("backend connection handler started")
        while not kill_ev.is_set():
            # Try to connect to the backend
            try:
                self._connect()
            except BastioBackendError as ex:
                self.close()
                self._logger.critical(ex.message)
                # TODO: Implement a more decent reconnection strategy
                time.sleep(5) # Sleep 5 seconds before retrial
                continue

            # Read a message from the wire, parse it, and push it to ingress queue(s)
            try:
                json_string = self._read_message()
                message = MessageParser.parse(json_string)
                self._put_ingress(message)
            except socket.timeout:
                pass # No messages are ready to be read
            except BastioNetstringError as ex:
                self._logger.critical(
                        "error parsing a Netstring message: {}".format(ex.message))
                self.close()
                continue
            except BastioMessageError as ex:
                self._logger.critical(
                        "error parsing a protocol message: {}".format(ex.message))
                self.close()
                continue
            except BastioEOFError:
                self._logger.critical("received EOF on channel")
                self.close()
                continue

            # Get an item from the egress queue(s) and send it to the backend
            try:
                message = self._get_egress(timeout=0.01) # 10ms
                if message == None:
                    # No message is available to send
                    continue
                self._write_message(message.to_json())
            except socket.timeout:
                # Too many un-ACK'd packets? Sliding window shut on our fingers?
                # We don't really know what happened, lets reschedule the last
                # message for retransmission anyway
                self._push_queue(self._tx, message)
            except BastioEOFError:
                # Message was not sent because channel was closed
                # re-push the message to the TX queue again and retry connection
                self._push_queue(self._tx, message)
                self.close()
                continue

    def _connect(self):
        """An idempotent method to connect to the backend."""
        try:
            if self._connected:
                return
            # Prepare host keys
            self._client = paramiko.SSHClient()
            hostkeys = self._client.get_host_keys()
            hostkey_server_name = self._make_hostkey_entry_name(self._backend_addr)
            hostkeys.add(hostkey_server_name, self._backend_hostkey.get_name(),
                    self._backend_hostkey)

            # Try to connect
            self._client.connect(hostname=self._backend_addr[0],
                    port=self._backend_addr[1], username=self._username,
                    pkey=self._agent_key, allow_agent=False,
                    look_for_keys=False)
            self._connected = True

            # Open session and establish the subsystem
            self._chan = self._invoke_bastio()
            self._logger.critical("connection established with the backend")
        except BastioBackendError:
            raise
        except paramiko.AuthenticationException:
            reraise(BastioBackendError, "authentication with backend failed")
        except paramiko.BadHostKeyException:
            reraise(BastioBackendError, "backend host key does not match")
        except socket.error as ex:
            reraise(BastioBackendError, ex.strerror.lower())
        except Exception:
            reraise(BastioBackendError)

    def _invoke_bastio(self):
        """Start a bastio subsystem on an already authenticated transport.

        :returns:
            A channel connected to the subsystem or None.
        """
        if not self.is_active():
            raise BastioBackendError("client is not connected")
        t = self._client.get_transport()
        chan = t.open_session()
        if not chan:
            raise BastioBackendError("opening a session with the backend failed")
        chan.settimeout(0.01) # 10ms
        chan.invoke_subsystem(self.Subsystem)
        return chan

    def _read_message(self):
        nets = Netstring(self._chan)
        return nets.recv()

    def _write_message(self, data):
        nets = Netstring.compose(data)
        remaining = len(nets)
        while remaining > 0:
            n = self._chan.send(nets)
            if n <= 0:
                raise BastioEOFError("channel closed")
            remaining -= n

    def _put_ingress(self, item):
        for endpoint in self._endpoints:
            endpoint.ingress.put(item)

    def _get_egress(self, timeout):
        for endpoint in self._endpoints:
            try:
                item = endpoint.egress.get_nowait()
                self._tx.put(item)
            except queue.Empty:
                pass
        try:
            return self._tx.get(timeout=timeout)
        except queue.Empty:
            return None

    @staticmethod
    def _catch_fail(failure):
        log = Logger()
        try:
            raise failure.exception, failure.message, failure.traceback
        except:
            msg = 'unexpected error occurred: {}'.format(failure.message)
            log.critical(msg, exc_info=True)

    @staticmethod
    def _make_hostkey_entry_name(addr):
        """We do the following to work around a paramiko inconsistency."""
        if addr[1] == paramiko.config.SSH_PORT:
            return addr[0]
        return '[{}]:{}'.format(*addr)




#!/usr/bin/env python
# Copyright 2013 Databracket LLC
# See LICENSE file for details.

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

# Make sure setuptools is installed
from distribute_setup import use_setuptools
use_setuptools()

import os
from setuptools import setup, find_packages
from bastio import __version__

def __filter_requires(filename):
    # Unnecessary packages for agent's normal operations
    unreqs = ['sphinx', 'theme', 'fabric']
    with open(filename, 'rb') as fd:
        reqs = map(lambda req: req.strip('\n'), fd.readlines())
    for unreq in unreqs:
        reqs = filter(lambda x: unreq not in x.lower(), reqs)
    return reqs

BASE_DIR = os.path.dirname(__file__)
README_PATH = os.path.join(BASE_DIR, 'README.rst')
REQS_PATH = os.path.join(BASE_DIR, 'requirements.txt')

classifiers = [
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Natural Language :: English',
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Intended Audience :: System Administrators',
    'Topic :: System :: Systems Administration',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    ]

setup(
        name = 'bastio-agent',
        version = __version__,
        description = 'Bastio agent to provision system accounts and SSH access',
        long_description = file(README_PATH).read(),
        author = 'Amr Ali',
        author_email = 'amr@databracket.com',
        maintainer = 'Amr Ali',
        maintainer_email = 'amr@databracket.com',
        url = 'https://bastio.com',
        install_requires = __filter_requires(REQS_PATH),
        entry_points = {
            'console_scripts': [
                'bastio-agent = bastio.cli:bastio_main',
                ],
            },
        packages = find_packages(),
        test_suite = 'bastio.load_test_suite',
        license = 'GPLv3+',
        platforms = 'Posix',
        classifiers = classifiers,
    )




# Copyright 2013 Databracket LLC
# See LICENSE file for details.

"""
:module: bastio.test.test_configs
:synopsis: Unit tests for the configs module.
:author: Amr Ali <amr@databracket.com>
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Databracket LLC"
__license__ = "GPLv3+"

import os
import unittest

from bastio.configs import GlobalConfigStore
from bastio.excepts import BastioConfigError

class TestGlobalConfigStore(unittest.TestCase):
    def setUp(self):
        self._conffile = 'TEST_CONFIG_FILE'
        confs = """[agent]
test_value = hello world
test_int = 434
[test]
test1_value = hello
test1_float = 44.3
        """
        file(self._conffile, 'wb').write(confs)

    def test_global_config_store(self):
        cfg = GlobalConfigStore()
        cfg.initialized = True
        self.assertEqual(cfg, GlobalConfigStore())
        self.assertTrue(cfg.initialized)
        self.assertTrue(GlobalConfigStore().initialized)
        self.assertEqual(cfg.initialized, GlobalConfigStore().initialized)

        # Test configuration file
        cfg.load(self._conffile)
        self.assertEqual(cfg.test_value, 'hello world')
        self.assertEqual(cfg.test_int, '434')
        self.assertEqual(cfg.get_agent_test_value, cfg.test_value)
        self.assertEqual(cfg.get_test_test1_value, 'hello')
        self.assertEqual(cfg.getfloat_test_test1_float, 44.3)

    def tearDown(self):
        os.unlink(self._conffile)

tests = [
        TestGlobalConfigStore,
        ]

